#ifndef AST_H
#define AST_H

#include <ostream>
#include <sstream>
#include <vector>
#include <map>

#include "llvm/DerivedTypes.h"
#include "llvm/LLVMContext.h"
#include "llvm/Module.h"
#include "llvm/Analysis/Verifier.h"
#include "llvm/Support/IRBuilder.h"

#include "llvm/Analysis/ProfileInfo.h"
#include "llvm/Support/raw_ostream.h"
#include "type_factory.h"

using namespace llvm;

Value* errorV(const char *str);
Function* errorF(const char *str);

extern Module* theModule;
extern TypeFactory* typeFactory;
extern IRBuilder<>* builder;

static std::map<std::string, Variable> namedValues;
static Value* object_struct_ptr;
static Function* llvm_malloc = NULL;
static Function* llvm_free = NULL;
static Function* currentFunction = NULL;
static const size_t WORD_SIZE = 8;
static LangType* currentClass = NULL;

void createMalloc();
void createFree();

class HasToString
{
    virtual std::string toString() const = 0;

public:
    friend std::ostream& operator<<(std::ostream& os, const HasToString& o)
    {
        os << o.toString();
        return os;
    }
};

template<typename T>
class CodeGenerator
{
public:
    virtual T* codegen() const = 0;
};

class ASTNode : public HasToString, public CodeGenerator<Value>
{
public:
    virtual ~ASTNode() {}
};

template<class ElementT, class CodegenT>
class NodeContainer
{
private:
    NodeContainer();
    ~NodeContainer();
    NodeContainer(const NodeContainer&);
    NodeContainer& operator=(const NodeContainer&);
};

template<class ElementT, class CodegenT>
class NodeContainer<ElementT*, CodegenT> :
      public CodeGenerator<CodegenT>, public HasToString
{
protected:
    typedef std::vector<ElementT*> Container;
    Container nodesList_;
public:
    typedef typename std::vector<ElementT*>::iterator iterator;
    typedef typename std::vector<ElementT*>::const_iterator const_iterator;

    iterator begin()
    {
        return nodesList_.begin();
    }

    const_iterator begin() const
    {
        return nodesList_.begin();
    }

    iterator end()
    {
        return nodesList_.end();
    }

    const_iterator end() const
    {
        return nodesList_.end();
    }

    void append(ElementT* node)
    {
        nodesList_.push_back(node);
    }

    void unshift(ElementT* node)
    {
        nodesList_.insert(nodesList_.begin(), node);
    }

    size_t size() const
    {
        return nodesList_.size();
    }

    const ElementT* operator[](size_t i) const
    {
        return nodesList_[i];
    }

    virtual ~NodeContainer()
    {
        for(typename Container::const_iterator it = nodesList_.begin();
            it != nodesList_.end(); ++it)
            delete *it;
    }

    virtual CodegenT* codegen() const
    {
        for(typename Container::const_iterator it = nodesList_.begin();
                it != nodesList_.end(); ++it)
            (*it)->codegen();
        return 0;
    }

    virtual std::string toString() const
    {
        std::ostringstream oss;
        oss << "(";
        for(typename Container::const_iterator it = nodesList_.begin(); it != nodesList_.end(); ++it)
            oss << **it << " ";
        oss << ")";
        return oss.str();
    }
};

class BlockASTNode : public NodeContainer<ASTNode*, BasicBlock>
{
public:
    virtual BasicBlock* codegen() const;
};

class UnitASTNode : public NodeContainer<ASTNode*, Value>
{
public:
    virtual std::string toString() const;
};

class NumberASTNode : public ASTNode
{
public:
    virtual std::string toString() const = 0;
    virtual Value* codegen() const = 0;
};

class IntNumberASTNode : public NumberASTNode
{
    const int val_;
public:
    IntNumberASTNode(int val) : val_(val) { };
    virtual std::string toString() const;
    virtual Value* codegen() const;
};

class DoubleNumberASTNode : public NumberASTNode
{
    const double val_;
public:
    DoubleNumberASTNode(double val) : val_(val) { };
    virtual std::string toString() const;
    virtual Value* codegen() const;
};

class VariableDeclASTNode : public ASTNode
{
protected:
    const std::string name_;
    const std::string type_;
    const int size_;
public:
    VariableDeclASTNode(
        const std::string& name,
        const std::string& type,
        const int size = 0
    ) : name_(name), type_(type), size_(size)
    {
    }

    Type* getType() const
    {
        if(size_ == 0)
            return typeFactory->getType(type_)->llvm_type;
        else
            return typeFactory->getType(type_)->llvm_ptr_type;
    }

    size_t getSize() const
    {
        size_t type_size = typeFactory->getType(type_)->size;
        return size_ == 0 ? type_size : size_ * type_size ;
    }

    std::string getLangType() const
    {
        return type_;
    }

    std::string getName() const
    {
        return name_;
    }

    virtual std::string toString() const;
    virtual AllocaInst* codegen() const;
};

class ArgumentASTNode : public VariableDeclASTNode
{
public:
    ArgumentASTNode (
        const std::string& name,
        const std::string& type,
        const int size = 0
    ) : VariableDeclASTNode(name, type, size)
    {
    }
    virtual AllocaInst* codegen() const;
};

class VariableDeleteASTNode : public ASTNode
{
    const std::string name_;

public:
    VariableDeleteASTNode(const std::string& name) : name_(name)
    {
    }

    virtual std::string toString() const;
    virtual Value* codegen() const;
};

class AbstractVariableASTNode : public ASTNode
{
public:
    virtual std::string toString() const = 0;
    virtual std::string getName() const = 0;
    virtual Value* codegen() const = 0; // rValue
    virtual Value* get() const = 0;     // lValue
};

class VariableASTNode : public AbstractVariableASTNode
{
    const std::string name_;
public:
    VariableASTNode(const std::string& name) : name_(name) {}

    virtual std::string toString() const;
    virtual Value* codegen() const;
    virtual Value* get() const;

    virtual std::string getName() const
    {
        return name_;
    }
};

class ArraySubscriptingASTNode : public AbstractVariableASTNode
{
    const AbstractVariableASTNode* var_;
    const ASTNode* index_;
public:
    ArraySubscriptingASTNode(
        const AbstractVariableASTNode* var,
        const ASTNode* index) :
        var_(var),
        index_(index)
    {
    }

    virtual std::string toString() const;
    virtual Value* codegen() const;
    virtual Value* get() const;
    virtual std::string getName() const
    {
        return var_->getName();
    }
};

typedef NodeContainer<VariableDeclASTNode*, Value> VariablesContainer;
typedef NodeContainer<ArgumentASTNode*, Value> ArgumentsContainer;
typedef NodeContainer<ASTNode*, Value> ASTNodeContainer;

class ReturnASTNode : public ASTNode
{
    const ASTNode* rhs_;
    public:
    ReturnASTNode(const ASTNode* rhs) : rhs_(rhs)
    {
    }

    virtual ~ReturnASTNode()
    {
        delete rhs_;
    }

    virtual std::string toString() const;
    virtual Value* codegen() const;
};

class UnaryOpASTNode : public ASTNode
{
    const std::string op_;
    const ASTNode* child_;
    public:
    UnaryOpASTNode(const std::string& op, const ASTNode* child) :
        op_(op),
        child_(child)
    {
    }

    virtual ~UnaryOpASTNode()
    {
        delete child_;
    }

    virtual std::string toString() const;
    virtual Value* codegen() const;
};

class BinaryOpASTNode : public ASTNode
{
protected:
    const std::string op_;
    const ASTNode* lhs_;
    const ASTNode* rhs_;
public:
    BinaryOpASTNode(
        const std::string& op, 
        const ASTNode* lhs,
        const ASTNode* rhs) :
        op_(op),
        lhs_(lhs),
        rhs_(rhs)
    {
    }

    virtual ~BinaryOpASTNode()
    {
        delete lhs_;
        delete rhs_;
    }

    virtual std::string toString() const;
    virtual Value* codegen() const;
};

class AssignmentASTNode : public BinaryOpASTNode
{
public:
    AssignmentASTNode(
        const std::string& op,
        const ASTNode* lhs,
        const ASTNode* rhs) :
        BinaryOpASTNode(op, lhs, rhs)
    {
    }
    virtual Value* codegen() const;
};

class FuncCallASTNode : public ASTNode
{
    const std::string functionName_;
    ASTNodeContainer* params_;

public:
    FuncCallASTNode(
        const std::string& functionName,
        ASTNodeContainer* params = new ASTNodeContainer()) :
        functionName_(functionName),
        params_(params)
    {
    }

    virtual ~FuncCallASTNode()
    {
        delete params_;
    }

    virtual std::string toString() const;
    virtual Value* codegen() const;
};

class FuncDefASTNode : public ASTNode
{
    const std::string functionName_;
    const std::string return_type_;
    const ArgumentsContainer* params_;
    bool extern_;
    const bool returns_array_;
    BlockASTNode* body_;
public:
    FuncDefASTNode(
        const std::string& functionName,
        const std::string& return_type,
        const bool returns_array = false,
        const ArgumentsContainer* params = new ArgumentsContainer()
    ) :
        functionName_(functionName),
        params_(params),
        return_type_(return_type),
        returns_array_(returns_array),
        extern_(false),
        body_(0)
    {
    }

    void makeExtern()
    {
        extern_ = true;
    }

    void addBody(BlockASTNode* body)
    {
        extern_ = false;
        body_ =body;
    }

    std::string getName() const
    {
        return functionName_;
    }

    size_t getNumParams() const
    {
        return params_->size();
    }

    Type* getType() const
    {
        if(returns_array_)
            return typeFactory->getType(return_type_)->llvm_ptr_type;
        else
            return typeFactory->getType(return_type_)->llvm_type;
    }

    void getArgs(std::vector<Type*>& args) const
    {
        for(std::size_t idx = 0; idx != params_->size(); ++idx)
            args.push_back((*params_)[idx]->getType());
    }

    virtual std::string toString() const;
    virtual Function* codegen() const;
};

typedef NodeContainer<FuncDefASTNode*, Value> FuncContainer;

class ConstructorDefASTNode : public ASTNode
{
    const std::string name_;
    const ArgumentsContainer* params_;
    BlockASTNode* body_;

public:
    ConstructorDefASTNode(
        const std::string& name,
        BlockASTNode* body,
        const ArgumentsContainer* params = new ArgumentsContainer()
    ) :
        name_(name),
        params_(params),
        body_(body)
    {
    }

    std::string getName() const
    {
        return name_;
    }

    virtual std::string toString() const;
    virtual Function* codegen() const;
};


class IfASTNode : public ASTNode
{
    const ASTNode* cond_;
    const BlockASTNode* if_body_;
    const BlockASTNode* else_body_;

public:
    IfASTNode( const ASTNode* cond,
            const BlockASTNode* if_body,
            const BlockASTNode* else_body = new BlockASTNode() ) :
        cond_(cond),
        if_body_(if_body),
        else_body_(else_body)
    {
    }

    virtual ~IfASTNode()
    {
        delete cond_;
        delete if_body_;
        delete else_body_;
    }

    virtual BasicBlock* codegen() const;
    virtual std::string toString() const;
};

class WhileASTNode : public ASTNode
{
    const ASTNode* cond_;
    const BlockASTNode* body_;

public:
    WhileASTNode(const ASTNode* cond,
            const BlockASTNode* body) :
        cond_(cond),
        body_(body)
    {
    }

    virtual ~WhileASTNode()
    {
        delete cond_;
        delete body_;
    }
    virtual BasicBlock* codegen() const;
    virtual std::string toString() const;
};

class ForASTNode : public ASTNode
{
    const ASTNodeContainer* begin_;
    const ASTNode* cond_;
    const ASTNodeContainer* step_;
    const BlockASTNode* body_;

public:
    ForASTNode(const ASTNodeContainer* begin,
            const ASTNode* cond,
            const ASTNodeContainer* step,
            const BlockASTNode* body) :
        begin_(begin),
        cond_(cond),
        step_(step),
        body_(body)
    {
    }

    virtual ~ForASTNode()
    {
        delete begin_;
        delete cond_;
        delete step_;
        delete body_;
    }
    virtual BasicBlock* codegen() const;
    virtual std::string toString() const;
};

class ClassBodyASTNode : public ASTNode
{
    friend class ClassDefASTNode;
    VariablesContainer fields_;
    FuncContainer methods_;
    FuncContainer virt_methods_;
    ConstructorDefASTNode* constructor_;

public:
    void addField(VariableDeclASTNode* field)
    {
        fields_.append(field);
    }

    void addMethod(FuncDefASTNode* method, bool is_virt = false)
    {
        if(is_virt) {
            virt_methods_.append(method);
        }
        else {
            methods_.append(method);
        }
    }

    void setConstructor(ConstructorDefASTNode* constrDef)
    {
        constructor_ = constrDef;
    }

    virtual std::string toString() const;
    virtual Value* codegen() const;
};

class ClassDefASTNode : public ASTNode
{
    const std::string name_;
    ClassBodyASTNode* body_;
public:
    ClassDefASTNode(
        const std::string& name,
        std::vector<std::string>* parents,
        ClassBodyASTNode* body
    ) : name_(name), body_(body)
    {
        int field_idx = 0;
        size_t type_size = 0;
        LangType lang_type;
        std::vector<Type*> elements;

        StructType* struct_type = StructType::create(
            getGlobalContext(), StringRef(name_)
        );

        if(parents->size() != 0)
        {
            LangType* parent_type = typeFactory->getType(parents->at(0));
            lang_type.vtable = parent_type->vtable;
        }

        if(body_->virt_methods_.size() != 0 &&
           lang_type.vtable.size() == 0)
        {
            elements.push_back(Type::getInt8Ty(getGlobalContext()));
            field_idx++;
            type_size++;
        }

        for(size_t i = 0; i < parents->size(); ++i, ++field_idx)
        {
            LangType* parent_type = typeFactory->getType(parents->at(i));
            elements.push_back(
                ((PointerType*) parent_type->llvm_type)->getElementType()
            );
            lang_type.castOffsets[parent_type] = type_size;

            for(std::map<LangType*, size_t>::const_iterator it =
                parent_type->castOffsets.begin();
                it != parent_type->castOffsets.end(); ++it) {
                lang_type.castOffsets[it->first] = it->second + type_size;
            }
            type_size += parent_type->size;

            lang_type.parents.push_back(parent_type);

            lang_type.parent_vtables.insert(
                parent_type->parent_vtables.begin(),
                parent_type->parent_vtables.end()
            );

            if(i != 0)
                lang_type.parent_vtables[parent_type] = parent_type->vtable;
        }

        for(VariablesContainer::iterator it = body_->fields_.begin();
            it != body_->fields_.end(); ++it, ++field_idx)
        {
            if((*it)->getLangType() != name_)
                elements.push_back((*it)->getType());
            else
                elements.push_back(PointerType::getUnqual(struct_type));
            std::string field_name = (*it)->getName();
            lang_type.fields[field_name] = ConstantInt::get(
                getGlobalContext(), APInt(32, field_idx)
            );
        }

        type_size += WORD_SIZE * field_idx;

        for(FuncContainer::iterator it = body_->methods_.begin();
            it != body_->methods_.end(); ++it)
        {
            std::ostringstream tmp_name;
            tmp_name << "__";
            tmp_name << name_;
            tmp_name << (*it)->getName();
            tmp_name << (*it)->getNumParams() + 1;

            lang_type.methods[(*it)->getName()] = tmp_name.str();
        }

        struct_type->setBody(ArrayRef<Type*>(elements));

        PointerType* struct_ptr = PointerType::get(struct_type, 0);
        PointerType* struct_ptr_ptr = PointerType::get(struct_ptr, 0);

        lang_type.type_name = name_;
        lang_type.llvm_type = struct_ptr;
        lang_type.llvm_ptr_type = struct_ptr_ptr;
        lang_type.size = type_size;

        typeFactory->addType(
            lang_type
        );

        fillVtable();
    }

    void fillVtable()
    {
        LangType* type = typeFactory->getType(name_);

        for(FuncContainer::iterator it = body_->virt_methods_.begin();
            it != body_->virt_methods_.end(); ++it)
        {
            std::string func_name = (*it)->getName();

            bool found = false;

            for(std::map<LangType*, vtable_t>::iterator parents_it =
                type->parent_vtables.begin();
                parents_it != type->parent_vtables.end(); ++parents_it)
            {
                if(parents_it->second.find(func_name) !=
                    parents_it->second.end())
                {
                    parents_it->second[func_name].class_owner = type;
                    found = true;
                }
            }

            vtable_t::iterator self_it = type->vtable.find(func_name);
            if(self_it != type->vtable.end())
            {
                type->vtable[func_name].class_owner = type;
                found = true;
            }

            if(! found)
            {
                std::vector<Type*>* args = new std::vector<Type*>();
                args->push_back(type->llvm_type);
                (*it)->getArgs(*args);

                VTEntry entry = VTEntry();

                entry.class_owner = type;
                entry.return_type = (*it)->getType();
                entry.offset = type->vtable.size();
                entry.base_class = type;
                type->vtable[func_name] = entry;
                type->vtable[func_name].args = args;
            }
        }
    }

    void createLLVMVtables() const
    {
        namedValues.clear();
        bool not_empty = false;

        LangType* type = typeFactory->getType(name_);

        std::string functionName;

        std::ostringstream tmp_name;
        tmp_name << "__";

        tmp_name << currentClass->type_name;
        tmp_name << "_vtables_constructor";

        functionName = tmp_name.str();

        FunctionType *ft;
        ft = FunctionType::get(
            Type::getVoidTy(getGlobalContext()),
            currentClass->llvm_type,
            false
        );
        Function *f = Function::Create(
            ft, Function::ExternalLinkage, functionName, theModule
        );

        currentFunction = f;

        BasicBlock* entry = BasicBlock::Create(
            getGlobalContext(),
            functionName + "_entry",
            currentFunction
        );
        builder->SetInsertPoint(entry);

        Function::arg_iterator it = f->arg_begin();

        it->setName("__this_" + currentClass->type_name);
        object_struct_ptr = it;

        GlobalVariable* llvm_vtable = createVtable(type, type->vtable);

        if(llvm_vtable != NULL)
        {
            not_empty = true;
            Value* vtable_ptr = builder->CreateBitCast(
                object_struct_ptr,
                PointerType::getUnqual(
                    PointerType::getUnqual(
                        PointerType::getUnqual(
                            Type::getInt8Ty(
                                getGlobalContext()
                            )
                        )
                    )
                )
            );

            Value* llvm_vtable_ptr = builder->CreateBitCast(
                llvm_vtable,
                PointerType::getUnqual(
                    PointerType::getUnqual(
                        Type::getInt8Ty(
                            getGlobalContext()
                        )
                    )
                )
            );

            builder->CreateStore(llvm_vtable_ptr, vtable_ptr);
        }

        for(std::map<LangType*, vtable_t>::iterator parents_it =
            type->parent_vtables.begin();
            parents_it != type->parent_vtables.end(); ++parents_it)
        {
            llvm_vtable = createVtable(
                parents_it->first, parents_it->second
            );

            if(llvm_vtable != NULL)
            {
                not_empty = true;
                Value* vtable_ptr = builder->CreateBitCast(
                    type->castTo(
                        object_struct_ptr,
                        parents_it->first
                    ),
                    PointerType::getUnqual(
                        PointerType::getUnqual(
                            PointerType::getUnqual(
                                Type::getInt8Ty(
                                    getGlobalContext()
                                )
                            )
                        )
                    )
                );

                Value* llvm_vtable_ptr = builder->CreateBitCast(
                    llvm_vtable,
                    PointerType::getUnqual(
                        PointerType::getUnqual(
                            Type::getInt8Ty(
                                getGlobalContext()
                            )
                        )
                    )
                );

                builder->CreateStore(llvm_vtable_ptr, vtable_ptr);
            }
        }

        builder->CreateRetVoid();
        if(not_empty)
        {
#ifndef MY_DEBUG
            verifyFunction(*f);
#endif
        }
        else
        {
            f->eraseFromParent();
        }
    }

    GlobalVariable* createVtable(
        LangType* vtableType, const vtable_t& vtable) const
    {
        if(vtable.size() == 0)
            return NULL;

        LangType* type = typeFactory->getType(name_);
        std::vector<Constant*> vtable_vect(vtable.size());

        for(vtable_t::const_iterator vmethod_it = vtable.begin();
            vmethod_it != vtable.end(); ++vmethod_it)
        {
            std::string functionName;
            std::ostringstream tmp_name;
            tmp_name << "__";
            tmp_name << vmethod_it->second.class_owner->type_name;
            tmp_name << vmethod_it->first;
            tmp_name << vmethod_it->second.args->size();

            functionName = tmp_name.str();
            Function* vmethod = theModule->getFunction(functionName);
            vtable_vect[vmethod_it->second.offset] =
                ConstantExpr::getCast(
                    Instruction::BitCast,
                    vmethod,
                    PointerType::get(
                        IntegerType::get(getGlobalContext(), 8), 0
                    )
                );
        }

        std::ostringstream vtable_name;
        vtable_name << "__";
        vtable_name << type->type_name;
        vtable_name << "_";
        vtable_name << vtableType->type_name;
        vtable_name << "_vtable";

        return new GlobalVariable(
            *theModule,
            ArrayType::get(
                Type::getInt8PtrTy(
                    getGlobalContext()
                ), vtable_vect.size()
            ),
            true,
            GlobalValue::ExternalLinkage,
            ConstantArray::get(
                ArrayType::get(
                    Type::getInt8PtrTy(
                        getGlobalContext()
                    ), vtable_vect.size()
                ),
                ArrayRef<Constant*> (vtable_vect)
            ),
            vtable_name.str()
        );
    }

    virtual std::string toString() const;
    virtual BasicBlock* codegen() const;

    VariablesContainer::iterator fields_begin()
    {
        return body_->fields_.begin();
    }

    VariablesContainer::const_iterator fields_begin() const
    {
        return body_->fields_.begin();
    }

    VariablesContainer::iterator fields_end()
    {
        return body_->fields_.end();
    }

    VariablesContainer::const_iterator fields_end() const
    {
        return body_->fields_.end();
    }
};

class MethodCallASTNode : public ASTNode
{
    AbstractVariableASTNode* object_;
    const std::string function_name_;
    ASTNodeContainer* params_;

public:
    MethodCallASTNode(
        AbstractVariableASTNode* object,
        const std::string& function_name,
        ASTNodeContainer* params
    ) : object_(object),
        function_name_(function_name),
        params_(params)
    {
    }

    virtual Value* codegen() const;
    Value* createVcall(
        LangType* object_type, const VTEntry& vtentry
    ) const;
    virtual std::string toString() const;
};

class ThisASTNode : public AbstractVariableASTNode
{
public:
    virtual std::string toString() const;

    virtual std::string getName() const
    {
        return "__this";
    }

    virtual Value* codegen() const; // rValue
    virtual Value* get() const;     // lValue
};

class FieldASTNode : public AbstractVariableASTNode
{
    const AbstractVariableASTNode* object_;
    const std::string field_;
public:
    FieldASTNode(
        AbstractVariableASTNode* object,
        const std::string& field
    ) : object_(object), field_(field)
    {
    }

    virtual std::string toString() const;
    virtual std::string getName() const
    {
        return object_->getName();
    }

    virtual Value* codegen() const;
    virtual Value* get() const;
};


class ConstructorASTNode : public VariableDeclASTNode
{
    ASTNodeContainer* params_;
public:
    ConstructorASTNode(
        const std::string& name,
        const std::string& type,
        ASTNodeContainer* params
    ) : VariableDeclASTNode(name, type, 0), params_(params)
    {
    }

    virtual AllocaInst* codegen() const;
};

#endif
