#include "ast.hpp"

BasicBlock* BlockASTNode::codegen() const
{
    BasicBlock *bb = BasicBlock::Create(getGlobalContext(), "bb", currentFunction);
    builder->SetInsertPoint(bb);
    for (Container::const_iterator it = nodesList_.begin(); it != nodesList_.end(); ++it)
        (*it)->codegen();
    return bb;
}

Value* IntNumberASTNode::codegen() const
{
    return ConstantInt::get(getGlobalContext(), APInt(64, val_));
}

Value* DoubleNumberASTNode::codegen() const
{
    return ConstantFP::get(getGlobalContext(), APFloat(val_));
}

Value* ThisASTNode::codegen() const
{
    if(currentClass == NULL)
    {
        return errorV("\"this\" outside of class definition.");
    }
    return object_struct_ptr;
}

Value* ThisASTNode::get() const
{
    return errorV("assignment to \"this\" is not supported");
}

Value* VariableASTNode::codegen() const
{
    return builder->CreateLoad(get());
}

Value* VariableASTNode::get() const
{
    if(namedValues.find(name_) != namedValues.end()) {
        return namedValues[name_].alloca_;
    }

    if(currentClass != NULL)
    {
        LangType* fieldClass = currentClass->classOfField(name_); 
        if(fieldClass != NULL)
        {
            Value* casted = currentClass->castTo(
                object_struct_ptr, fieldClass);
            std::vector<Value*> tmp;           
            tmp.push_back(ConstantInt::get(getGlobalContext(), APInt(64, 0)));
            tmp.push_back(fieldClass->fields[name_]);
            return builder->CreateGEP(casted, ArrayRef<Value*>(tmp)); 
        }
    }

    return errorV("Unknown variable name");
}

AllocaInst* ArgumentASTNode::codegen() const
{
    return builder->CreateAlloca(getType(), 0, name_ + "_alloca");
}

Value* ArraySubscriptingASTNode::codegen() const
{
    return builder->CreateLoad(get());
}

Value* ArraySubscriptingASTNode::get() const
{
    Value* idx = index_->codegen();
    Value* ptr = builder->CreateLoad(var_->get());
    return builder->CreateGEP(ptr, idx);
}

Value* FieldASTNode::codegen() const
{
    return builder->CreateLoad(get());
}

Value* FieldASTNode::get() const
{
    Value* ptr = object_->codegen();
    const std::string name = object_->getName();

    if(namedValues.find(name) != namedValues.end()) {
        LangType* type = namedValues[name].type_name;

        LangType* fieldClass = type->classOfField(field_); 
        if(fieldClass != NULL)
        {
            Value* casted = type->castTo(ptr, fieldClass);
            std::vector<Value*> tmp;           
            tmp.push_back(ConstantInt::get(getGlobalContext(), APInt(64, 0)));
            tmp.push_back(fieldClass->fields[field_]);
            return builder->CreateGEP(casted, ArrayRef<Value*>(tmp)); 
        }
    }

    return errorV("Unknown variable name");
}

Value* VariableDeleteASTNode::codegen() const
{
    if(llvm_free == NULL)
        createFree();

    Value* ptr = builder->CreateLoad(namedValues[name_].alloca_);

    if(ptr == NULL)
        return errorV("Unknown variable name");

    builder->CreateCall(
        llvm_free,
        builder->CreateBitCast(
            ptr,
            Type::getInt8PtrTy(getGlobalContext())
        )
    );

    return NULL;
}

AllocaInst* VariableDeclASTNode::codegen() const
{
    Type* llvmType = getType();
    LangType* type = typeFactory->getType(type_);

    if(llvmType == NULL)
        return NULL;

    if(! type->llvm_type->isPointerTy() && size_ == 0) 
    {
        AllocaInst* alloca = builder->CreateAlloca(
            llvmType, 0, name_ + "_alloca"
        );

        namedValues[name_] = Variable();
        namedValues[name_].alloca_ = alloca;
        namedValues[name_].type_name = typeFactory->getType(type_);

        return alloca;
    }
    else {
        if(llvm_malloc == 0)
            createMalloc();
        AllocaInst* ptr = builder->CreateAlloca(
            llvmType, 0, name_ + "_ptr_alloca"
        );

        builder->CreateStore(
            builder->CreateBitCast(
                builder->CreateCall(
                    llvm_malloc,
                    ConstantInt::get(
                        getGlobalContext(), 
                        APInt(64, getSize())
                    )
                ),
                llvmType
            ),
            ptr
        );
        namedValues[name_] = Variable();
        namedValues[name_].alloca_ = ptr;
        namedValues[name_].type_name = typeFactory->getType(type_);

        std::ostringstream vt_constr_name;
        vt_constr_name << "__";
        
        vt_constr_name << type->type_name;
        vt_constr_name << "_vtables_constructor";

        Function* vt_constr = theModule->getFunction(vt_constr_name.str());

        if(vt_constr != NULL)
        {
            if(size_ == 0) {
                builder->CreateCall(
                    vt_constr, 
                    builder->CreateLoad(ptr)
                );
            }
            else
            {
                for(size_t i = 0; i < size_; ++i)
                {
                    std::vector<Value*> tmp;           
                    tmp.push_back(
                        ConstantInt::get(
                            getGlobalContext(), 
                            APInt(64, 0)
                        )
                    );
                    tmp.push_back(
                        ConstantInt::get(
                            getGlobalContext(), 
                            APInt(64, i)
                        )
                    );
                    builder->CreateCall(
                        vt_constr, 
                        builder->CreateLoad(
                            builder->CreateGEP(
                                ptr, 
                                ArrayRef<Value*>(tmp)
                            )
                        )
                    );
                }
            }
        }

        return ptr;
    }
}

Function* ConstructorDefASTNode::codegen() const
{
    namedValues.clear();

    std::vector<Type*> llvm_params;

    llvm_params.push_back(currentClass->llvm_type);

    for(std::size_t idx = 0; idx != params_->size(); ++idx)
        llvm_params.push_back((*params_)[idx]->getType());

    std::string functionName;

    std::ostringstream tmp_name;
    tmp_name << "__";
    
    tmp_name << currentClass->type_name;
    tmp_name << "constructor";
    tmp_name << params_->size() + 1;

    functionName = tmp_name.str();

    FunctionType *ft;
    ft = FunctionType::get(Type::getVoidTy(getGlobalContext()), llvm_params, false);
    Function *f = Function::Create(ft, Function::ExternalLinkage, functionName, theModule);

    currentFunction = f;

    BasicBlock* entry = BasicBlock::Create(
        getGlobalContext(), functionName + "_entry", currentFunction
    );
    builder->SetInsertPoint(entry);

    Function::arg_iterator it = f->arg_begin();

    it->setName("__this_" + currentClass->type_name);
    object_struct_ptr = it;
    ++it;

    for (std::size_t idx = 0; idx != params_->size(); ++it, ++idx) {
        std::string paramName = (*params_)[idx]->getName();
        it->setName(paramName);
        namedValues[paramName] = Variable();
        namedValues[paramName].alloca_ = (*params_)[idx]->codegen();
        namedValues[paramName].type_name = 
            typeFactory->getType((*params_)[idx]->getLangType());
        builder->CreateStore(it, namedValues[paramName].alloca_);
    }

    if (f == 0) return 0;

    if (BasicBlock* bb = body_->codegen()) {
        builder->SetInsertPoint(entry);
        builder->CreateBr(bb);
        builder->SetInsertPoint(bb);
        builder->CreateRetVoid();
#ifndef MY_DEBUG
        verifyFunction(*f);
#endif
        return f;
    } else {
        f->eraseFromParent();
        return 0;
    }
}

AllocaInst* ConstructorASTNode::codegen() const
{
    AllocaInst* ret = VariableDeclASTNode::codegen(); 

    std::ostringstream tmp_name;
    tmp_name << "__";
    
    tmp_name << type_;
    tmp_name << "constructor";
    tmp_name << params_->size() + 1;

    Function* callee_f = theModule->getFunction(tmp_name.str());

    if(callee_f == 0) {
        return ret;
    }

    std::vector<Value*> argsV;
    argsV.push_back(builder->CreateLoad(ret));

    for (size_t i = 0, e = params_->size(); i != e; ++i) {
        argsV.push_back((*params_)[i]->codegen());
        if (argsV.back() == 0) return 0;
    }

    builder->CreateCall(callee_f, argsV);

    return ret;
}

Value* ReturnASTNode::codegen() const
{
    return builder->CreateRet(rhs_->codegen());
}

Value* UnaryOpASTNode::codegen() const
{
    Value *c = child_->codegen();
    if(c == 0) return 0;

    if(op_ == "!")
        return builder->CreateNeg(c);
    else
        return errorV("invalid binary operator");
}


Value* binaryOpIntegerCodegen(Value* l, Value* r, const std::string& op)
{
    if (op == "+")
        return builder->CreateAdd(l, r);
    else if (op == "-")
        return builder->CreateSub(l, r);
    else if (op == "*")
        return builder->CreateMul(l, r);
    else if (op == "/")
        return builder->CreateSDiv(l, r);
    else if (op == ">")
        return builder->CreateICmpSGT(l, r);
    else if (op == ">=")
        return builder->CreateICmpSGE(l, r);
    else if (op == "<")
        return builder->CreateICmpSLT(l, r);
    else if (op == "<=")
        return builder->CreateICmpSLE(l, r);
    else if (op == "==")
        return builder->CreateICmpEQ(l, r);
    else if (op == "!=")
        return builder->CreateICmpNE(l, r);
    else
        return errorV("invalid binary operator");
}

Value* binaryOpFloatingCodegen(Value* l, Value* r, const std::string& op)
{
    if (op == "+")
        return builder->CreateFAdd(l, r);
    else if (op == "-")
        return builder->CreateFSub(l, r);
    else if (op == "*")
        return builder->CreateFMul(l, r);
    else if (op == "/")
        return builder->CreateFDiv(l, r);
    else if (op == ">")
        return builder->CreateFCmpOGT(l, r);
    else if (op == ">=")
        return builder->CreateFCmpOGE(l, r);
    else if (op == "<")
        return builder->CreateFCmpOLT(l, r);
    else if (op == "<=")
        return builder->CreateFCmpOLE(l, r);
    else if (op == "==")
        return builder->CreateFCmpOEQ(l, r);
    else if (op == "!=")
        return builder->CreateFCmpONE(l, r);
    else
        return errorV("invalid binary operator");
}

Value* BinaryOpASTNode::codegen() const
{
    Value *l = lhs_->codegen();
    Value *r = rhs_->codegen();
    if (l == 0 || r == 0) return 0;
    
    Type* lt = l->getType();
    Type* rt = r->getType();

    if(lt->isIntegerTy() && rt->isIntegerTy())
        return binaryOpIntegerCodegen(l, r, op_);
    else if(lt->isFloatingPointTy() && rt->isFloatingPointTy() )
        return binaryOpFloatingCodegen(l, r, op_);
    else if(lt->isIntegerTy() && rt->isFloatingPointTy())
        return binaryOpFloatingCodegen(
            builder->CreateIntCast(l, rt, true), r, op_
        );
    else if(lt->isFloatingPointTy() && rt->isIntegerTy())
        return binaryOpFloatingCodegen(
            l, builder->CreateIntCast(r, lt, true), op_
        );
    else 
        return errorV("Bad operand type for binary operation");
}

Value* AssignmentASTNode::codegen() const
{
    const AbstractVariableASTNode* lval =
        dynamic_cast<const AbstractVariableASTNode*>(lhs_);

    if(lval == NULL)
    {
        return errorV("Assignment to non-variable type");
    }

    Value *l = lval->get();
    Value *r = rhs_->codegen();

    Type* lt = l->getType();
    Type* rt = r->getType();

    if(lt->isIntegerTy() && rt->isFloatingPointTy())
        return builder->CreateStore(builder->CreateFPCast(r, lt), l);
    else if(lt->isFloatingPointTy() && rt->isIntegerTy())
        return builder->CreateStore(
            l, builder->CreateIntCast(r, lt, true)
        );
    else if(rt->isPointerTy())
    {
        Type* elem_type = ((PointerType*) rt)->getElementType();
        if(elem_type->isStructTy())
        {
            std::string tn = 
                ((StructType*) elem_type)->getName().str();
            LangType* rlt = typeFactory->getType(tn);
            LangType* llt = namedValues[lval->getName()].type_name;

            return builder->CreateStore(rlt->castTo(r, llt), l);
        }
    }
    
    return builder->CreateStore(r, l);
}

Value* FuncCallASTNode::codegen() const
{
    if(currentClass != NULL)
    {
        LangType* meth_class =
            currentClass->classOfMethod(functionName_); 

        if(meth_class != NULL)
        {
            return (
                new MethodCallASTNode(
                    new ThisASTNode(), 
                    functionName_, 
                    params_
                )
            )->codegen();
        }
    }

    std::ostringstream tmp_name;
    tmp_name << "__";
    tmp_name << functionName_;
    tmp_name << params_->size();

    std::string functionName = tmp_name.str();
    Function* calleeF = theModule->getFunction(functionName);
    if(calleeF == 0) {
        calleeF = theModule->getFunction(functionName_);
        if (calleeF == 0)
            return errorV("Unknown function referenced");
    }

    if (calleeF->arg_size() != params_->size())
        return errorV("Incorrect # of arguments passed");

    std::vector<Value*> argsV;
    for (size_t i = 0, e = params_->size(); i != e; ++i) {
        argsV.push_back((*params_)[i]->codegen());
        if (argsV.back() == 0) return 0;
    }

    return builder->CreateCall(calleeF, argsV);
}

Value* MethodCallASTNode::codegen() const
{
    LangType* object_type;
    std::string object_name = object_->getName();

    if(object_name != "__this")
    {
        if(namedValues.find(object_name) == namedValues.end())
            return errorV("Unknown object");

        object_type = namedValues[object_->getName()].type_name;
    }
    else
    {
        object_type = currentClass;
    }

    vtable_t::const_iterator it = 
        object_type->vtable.find(function_name_);

    if(it != object_type->vtable.end())
    {
        return createVcall(object_type, it->second);
    }

    for(std::map<LangType*, vtable_t>::const_iterator parent_it =  
        object_type->parent_vtables.begin();
        parent_it != object_type->parent_vtables.end();
        ++parent_it)
    {
        vtable_t::const_iterator vfunct_it = 
           parent_it->second.find(function_name_);

        if(vfunct_it != parent_it->second.end())
        {
            return createVcall(object_type, vfunct_it->second);
            break;
        }
    }

    LangType* meth_class = object_type->classOfMethod(function_name_); 

    if(meth_class == NULL)
        return errorV("Unknown function referenced");

    std::string function_name = meth_class->methods[function_name_]; 
    Function* callee_f = theModule->getFunction(function_name);

    if(callee_f == 0) {
        return errorV("Unknown function referenced");
    }

    if (callee_f->arg_size() != params_->size() + 1)
        return errorV("Incorrect # arguments passed");

    std::vector<Value*> argsV;
    Value* object_ptr = object_->codegen();
    Value* casted = object_type->castTo(object_ptr, meth_class);
    argsV.push_back(casted);

    for (size_t i = 0, e = params_->size(); i != e; ++i) {
        argsV.push_back((*params_)[i]->codegen());
        if (argsV.back() == 0) return 0;
    }

    return builder->CreateCall(callee_f, argsV);
}


Value* MethodCallASTNode::createVcall(
    LangType* object_type, const VTEntry& vtentry) const
{
    LangType* meth_class = vtentry.base_class;
    size_t offset = vtentry.offset;
    std::vector<Value*> argsV;
    Value* object_ptr = object_->codegen();

    Value* casted = object_type->castTo(object_ptr, meth_class);
    argsV.push_back(casted);

    for (size_t i = 0, e = params_->size(); i != e; ++i) {
        argsV.push_back((*params_)[i]->codegen());
        if (argsV.back() == 0) return 0;
    }

    Value* vtable_ptr = builder->CreateBitCast(
        casted, 
        PointerType::getUnqual(
            PointerType::getUnqual(
                PointerType::getUnqual(
                    FunctionType::get(
                        vtentry.return_type, 
                        ArrayRef<Type*>(*vtentry.args), 
                        false
                    )
                )
            )
        )
    );

    Value* func_ptr = builder->CreateLoad(
        builder->CreateGEP(
            builder->CreateLoad(vtable_ptr), 
            ConstantInt::get(
                getGlobalContext(), 
                APInt(64, offset)
            )
        )
    );

    return builder->CreateCall(func_ptr, argsV);
}

Function* FuncDefASTNode::codegen() const
{
    namedValues.clear();

    std::vector<Type*> llvm_params;
    LangType* vmethod_def_class = NULL;

    if(currentClass != NULL) {
        vtable_t::const_iterator self_it = 
            currentClass->vtable.find(functionName_);

        if(self_it != currentClass->vtable.end())
        {
            vmethod_def_class = self_it->second.base_class;
        }
        else
        {
            for(std::map<LangType*, vtable_t>::const_iterator parent_it =  
                currentClass->parent_vtables.begin();
                parent_it != currentClass->parent_vtables.end();
                ++parent_it)
            {
                vtable_t::const_iterator vfunct_it = 
                parent_it->second.find(functionName_);

                if(vfunct_it != parent_it->second.end())
                {
                    vmethod_def_class = vfunct_it->second.base_class;
                    break;
                }
            }
        }

        if(vmethod_def_class != NULL)
        {
            llvm_params.push_back(vmethod_def_class->llvm_type);
        }
        else
        {
            llvm_params.push_back(
                currentClass->llvm_type
            );
        }
    }

    for(std::size_t idx = 0; idx != params_->size(); ++idx)
        llvm_params.push_back((*params_)[idx]->getType());

    Type* return_type = getType();

    if(return_type == NULL) {
        errorF("Wrong return type");
        return 0;
    }

    std::string functionName;
    if(functionName_ != "main") {
        std::ostringstream tmp_name;
        tmp_name << "__";
        
        if(currentClass != NULL) {
            tmp_name << currentClass->type_name;
        }

        tmp_name << functionName_;

        if(currentClass != NULL)
            tmp_name << params_->size() + 1;
        else 
            tmp_name << params_->size();

        functionName = tmp_name.str();
    }
    else {
        functionName = functionName_;
    }

    FunctionType *ft;
    ft = FunctionType::get(return_type, llvm_params, false);
    Function *f = Function::Create(ft, Function::ExternalLinkage, functionName, theModule);

    currentFunction = f;

    if (f->getName() != functionName) {
        f->eraseFromParent();
        f = theModule->getFunction(functionName);
        errorF("redefinition of function");
        return 0;
    }

    BasicBlock* entry = BasicBlock::Create(
        getGlobalContext(), functionName + "_entry", currentFunction
    );
    builder->SetInsertPoint(entry);

    Function::arg_iterator it = f->arg_begin();

    if(currentClass != NULL) {
        object_struct_ptr = it;

        if(vmethod_def_class != NULL)
        {
            it->setName("__this_" + vmethod_def_class->type_name);
            object_struct_ptr = currentClass->downcast(
                object_struct_ptr, 
                vmethod_def_class
            );
        }
        else
        {
            it->setName("__this_" + currentClass->type_name);
        }
        ++it;
    }

    for (std::size_t idx = 0; idx != params_->size(); ++it, ++idx) {
        std::string paramName = (*params_)[idx]->getName();
        it->setName(paramName);
        namedValues[paramName] = Variable();
        namedValues[paramName].alloca_ = (*params_)[idx]->codegen();
        namedValues[paramName].type_name = 
            typeFactory->getType((*params_)[idx]->getLangType());
        builder->CreateStore(it, namedValues[paramName].alloca_);
    }

    if (f == 0) return 0;

    if (BasicBlock* bb = body_->codegen()) {
        builder->SetInsertPoint(entry);
        builder->CreateBr(bb);
        builder->SetInsertPoint(bb);
#ifndef MY_DEBUG
        verifyFunction(*f);
#endif
        return f;
    } else {
        f->eraseFromParent();
        return 0;
    }
}

BasicBlock* IfASTNode::codegen() const
{
    BasicBlock* current = builder->GetInsertBlock();
    BasicBlock* if_block = if_body_->codegen();
    BasicBlock* if_block_end = builder->GetInsertBlock();
    BasicBlock* else_block = else_body_->codegen();
    BasicBlock* else_block_end = builder->GetInsertBlock();
    BasicBlock* if_end = BasicBlock::Create(getGlobalContext(), "if_end", currentFunction);

    builder->SetInsertPoint(current);

    builder->CreateCondBr(
            cond_->codegen(),
            if_block,
            else_block
    );

    builder->SetInsertPoint(if_block_end);
    builder->CreateBr(if_end);

    builder->SetInsertPoint(else_block_end);
    builder->CreateBr(if_end);

    builder->SetInsertPoint(if_end);

    return if_end;
}

BasicBlock* WhileASTNode::codegen() const
{
    BasicBlock* while_cond = BasicBlock::Create(getGlobalContext(), "while_cond", currentFunction);
    builder->CreateBr(while_cond);
    BasicBlock* body_block = body_->codegen();
    BasicBlock* while_end = BasicBlock::Create(getGlobalContext(), "while_end", currentFunction);
    builder->CreateBr(while_cond);
    builder->SetInsertPoint(while_cond);

    builder->CreateCondBr(
        cond_->codegen(),
        body_block,
        while_end
    );

    builder->SetInsertPoint(while_end);
    return while_end;
}

BasicBlock* ForASTNode::codegen() const
{
    BasicBlock* for_cond = BasicBlock::Create(getGlobalContext(), "for_cond", currentFunction);
    BasicBlock* for_end = BasicBlock::Create(getGlobalContext(), "for_end", currentFunction);

    begin_->codegen();
    builder->CreateBr(for_cond);

    BasicBlock* body = body_->codegen();
    step_->codegen();
    builder->CreateBr(for_cond);

    builder->SetInsertPoint(for_cond);
    builder->CreateCondBr(
        cond_->codegen(),
        body,
        for_end
    );
    builder->SetInsertPoint(for_end);
    return for_end;
}

BasicBlock* ClassDefASTNode::codegen() const
{
    currentClass = typeFactory->getType(name_);

    body_->methods_.codegen();
    body_->virt_methods_.codegen();
    createLLVMVtables();

    if(body_->constructor_ != NULL)
        body_->constructor_->codegen();

    currentClass = NULL;
    return NULL;
}

Value* ClassBodyASTNode::codegen() const
{
    return NULL;
}

void createMalloc()
{
    ArrayRef<Type*> params(Type::getInt64Ty(getGlobalContext()));
    llvm_malloc = Function::Create(
            FunctionType::get(
                Type::getInt8PtrTy(getGlobalContext()),
                params,
                false
                ), Function::ExternalLinkage, "malloc", theModule
            );
    verifyFunction(*llvm_malloc);
}

void createFree()
{
    ArrayRef<Type*> params(Type::getInt8PtrTy(getGlobalContext()));
    llvm_free = Function::Create(
            FunctionType::get(
                Type::getVoidTy(getGlobalContext()),
                params,
                false
                ), Function::ExternalLinkage, "free", theModule
            );
    verifyFunction(*llvm_free);
}
