%skeleton "lalr1.cc"   
%defines
%define parser_class_name "Parser"

%code requires {
#include <string>
#include "ast.hpp"
class Driver;
}

%parse-param { Driver& driver }
%lex-param   { Driver& driver }

%debug
%error-verbose

%union {
    double dval;
    int ival;
    std::string* sval;
    ArgumentsContainer* params;
    ArgumentASTNode* arg;
    ASTNodeContainer* nodes;
    std::vector<std::string>* strs;
    ASTNode* ast;
    NumberASTNode* num;
    AbstractVariableASTNode* var;
    VariableDeclASTNode* decl;
    BinaryOpASTNode* binOp;
    ClassDefASTNode* classDef;
    ClassBodyASTNode* classBody;
    FuncCallASTNode* call;
    ReturnASTNode* ret;
    FuncDefASTNode* func;
    ConstructorDefASTNode* constrDef;
    BlockASTNode* block;
    UnitASTNode* unit;
    ForASTNode*  for_node;
    IfASTNode*  if_node;
    WhileASTNode*  while_node;
}

%code {
#include "driver.hpp"
}

%token END 0 "end of file"
%token LT
%token LE
%token GE
%token GT
%token EQ
%token NE
%token NOT
%token PLUS
%token MINUS
%token TIMES
%token SLASH
%token LPAREN
%token RPAREN
%token BLOCK_BEGINS
%token BLOCK_ENDS
%token ARROW
%token ASSIGN
%token COMMA
%token CLASS
%token SEMICOLON
%token COLON
%token DEF
%token EXTERN
%token RETURN
%token VIRT
%token DELETE
%token WHILE
%token IF
%token ELSE
%token OF
%token FOR
%token LBRACKET
%token RBRACKET
%token <sval> IDENTIFIER "identifier"
%token <dval> DOUBLENUMBER "double number"
%token <ival> INTNUMBER "int number"
%type <var> variable
%type <ret> return;
%type <ast> expr
%type <ast> var_delete
%type <decl> var_decl
%type <call> func_call
%type <for_node> for_statement
%type <while_node> while_statement
%type <if_node> if_statement
%type <func> func_prototype
%type <func> func_definition
%type <block> code_block
%type <block> statements
%type <classDef> class_definition
%type <constrDef> constructor_def
%type <classBody> class_body
%type <nodes>  expressions
%type <arg> param
%type <params> prototype_params
%type <unit> unit
%type <unit> program
%type <strs> parents
%type <strs> parents_
%%

%right ASSIGN;
%left EQ NE;
%left LT LE GT GE;
%left PLUS MINUS;
%left TIMES SLASH;
%right NOT;
%left ARROW;
%left LBRACKET;

%start program;

program
    : unit { $$ = $1; driver.result = $$; driver.result->codegen(); }
    ;

unit
    : /* Nothing */        { $$ = new UnitASTNode(); }
    | unit class_definition { $$ = $1; $$->append($2); }
    | unit func_definition  { $$ = $1; $$->append($2); }
    ;

class_definition 
    : CLASS IDENTIFIER parents BLOCK_BEGINS class_body BLOCK_ENDS 
      { $$ = new ClassDefASTNode(*$2, $3, $5); }
    ;

parents
    : /* Nothing */  { $$ = new std::vector<std::string>(); }
    | COLON parents_ { $$ = $2; }
    ;

parents_
    : IDENTIFIER 
    { $$ = new std::vector<std::string>(); $$->push_back(*$1);}
    | parents_ COMMA IDENTIFIER 
    { $$ = $1; $$->push_back(*$3); }
    ;

class_body
    : /* Nothing */ 
      { $$ = new ClassBodyASTNode(); }
    | class_body var_decl SEMICOLON 
      { $$ = $1; $$->addField($2); }  
    | class_body constructor_def
      { $$ = $1; $$->setConstructor($2); }
    | class_body func_definition
      { $$ = $1; $$->addMethod($2); }  
    | class_body VIRT func_definition
      { $$ = $1; $$->addMethod($3, true); }  
    ;

code_block
    : BLOCK_BEGINS statements BLOCK_ENDS { $$ = $2; }
    ;

statements
    : /* Nothing */                       { $$ = new BlockASTNode(); }
    | statements var_decl SEMICOLON       { $$ = $1; $$->append($2); }
    | statements var_delete SEMICOLON     { $$ = $1; $$->append($2); }
    | statements expr SEMICOLON           { $$ = $1; $$->append($2); }
    | statements return SEMICOLON         { $$ = $1; $$->append($2); }
    | statements for_statement            { $$ = $1; $$->append($2); }
    | statements if_statement             { $$ = $1; $$->append($2); }
    | statements while_statement          { $$ = $1; $$->append($2); }
    ;

for_statement
    : FOR LPAREN expressions SEMICOLON expr SEMICOLON expressions RPAREN code_block
      { $$ = new ForASTNode($3, $5, $7, $9); }
    ;

if_statement
    : IF LPAREN expr RPAREN code_block ELSE code_block
      { $$ = new IfASTNode($3, $5, $7); }
    | IF LPAREN expr RPAREN code_block 
      { $$ = new IfASTNode($3, $5); }
    ;

while_statement
    : WHILE LPAREN expr RPAREN code_block
      { $$ = new WhileASTNode($3, $5); }
    ;

var_decl
    : IDENTIFIER OF IDENTIFIER LBRACKET INTNUMBER RBRACKET 
    { $$ = new VariableDeclASTNode(*$1, *$3, $5); }
    | IDENTIFIER OF IDENTIFIER 
    { $$ = new VariableDeclASTNode(*$1, *$3); }
    | IDENTIFIER OF IDENTIFIER LPAREN expressions RPAREN
    { $$ = new ConstructorASTNode(*$1, *$3, $5); }
    ;

var_delete
    : DELETE IDENTIFIER { $$ = new VariableDeleteASTNode(*$2); }

expr
    : NOT expr           
    { $$ = new UnaryOpASTNode("!", $2); }
    | expr LT expr       
    { $$ = new BinaryOpASTNode("<", $1, $3); }
    | expr LE expr       
    { $$ = new BinaryOpASTNode("<=", $1, $3); }
    | expr GT expr       
    { $$ = new BinaryOpASTNode(">", $1, $3); }
    | expr GE expr       
    { $$ = new BinaryOpASTNode(">=", $1, $3); }
    | expr EQ expr       
    { $$ = new BinaryOpASTNode("==", $1, $3); }
    | expr NE expr       
    { $$ = new BinaryOpASTNode("!=", $1, $3); }
    | expr PLUS expr     
    { $$ = new BinaryOpASTNode("+", $1, $3); }
    | expr MINUS expr    
    { $$ = new BinaryOpASTNode("-", $1, $3); }
    | expr TIMES expr    
    { $$ = new BinaryOpASTNode("*", $1, $3); }
    | expr SLASH expr    
    { $$ = new BinaryOpASTNode("/", $1, $3); }
    | expr ASSIGN expr 
    { $$ = new AssignmentASTNode("=", $1, $3); }
    | INTNUMBER          
    { $$ = new IntNumberASTNode($1); }
    | DOUBLENUMBER       
    { $$ = new DoubleNumberASTNode($1); }
    | LPAREN expr RPAREN 
    { $$ = $2; }
    | variable ARROW IDENTIFIER LPAREN expressions RPAREN
    { $$ = new MethodCallASTNode($1, *$3, $5); }
    | func_call          
    { $$ = $1; }
    | variable           
    { $$ = $1; }
    ;

return
    : RETURN expr { $$ = new ReturnASTNode($2); }
    ;

variable
    : variable LBRACKET expr RBRACKET   
    { $$ = new ArraySubscriptingASTNode($1, $3); }
    | variable ARROW IDENTIFIER         
    { $$ = new FieldASTNode($1, *$3) ; }
    | IDENTIFIER                        
    { $$ = new VariableASTNode(*$1); }
    ;

func_call
    : IDENTIFIER LPAREN expressions RPAREN { $$ = new FuncCallASTNode(*$1, $3); }
    ;

expressions
    : /* */                  
    { $$ = new ASTNodeContainer(); }
    | expr                   
    { $$ = new ASTNodeContainer(); $$->append($1); }
    | expressions COMMA expr 
    { $$ = $1; $$->append($3); }
    ;

func_prototype
    : IDENTIFIER LPAREN RPAREN OF IDENTIFIER 
      { $$ = new FuncDefASTNode(*$1, *$5); }
    | IDENTIFIER LPAREN RPAREN OF IDENTIFIER LBRACKET INTNUMBER RBRACKET 
      { $$ = new FuncDefASTNode(*$1, *$5, true); }
    | IDENTIFIER LPAREN prototype_params RPAREN OF IDENTIFIER 
      { $$ = new FuncDefASTNode(*$1, *$6, false, $3); }
    | IDENTIFIER LPAREN prototype_params RPAREN OF IDENTIFIER LBRACKET INTNUMBER RBRACKET
      { $$ = new FuncDefASTNode(*$1, *$6, true, $3); }
    ;

prototype_params
    : param
      { $$ = new ArgumentsContainer(); $$->append($1); }
    | prototype_params COMMA param 
      { $$ = $1; $$->append($3); }
    ;

param
    : IDENTIFIER OF IDENTIFIER LBRACKET INTNUMBER RBRACKET 
    { $$ = new ArgumentASTNode(*$1, *$3, $5); }
    | IDENTIFIER OF IDENTIFIER 
    { $$ = new ArgumentASTNode(*$1, *$3); }
    ;

func_definition
    : DEF func_prototype code_block   { $$ = $2; $$->addBody($3); }
    | EXTERN func_prototype SEMICOLON { $$ = $2; $$->makeExtern(); }
    ;

constructor_def
    : DEF IDENTIFIER LPAREN RPAREN code_block
      { $$ = new ConstructorDefASTNode(*$2, $5); }
    | DEF IDENTIFIER LPAREN prototype_params RPAREN code_block
      { $$ = new ConstructorDefASTNode(*$2, $6, $4); }
    ;

%%

void yy::Parser::error(const yy::location& l, const std::string& m)
{
    driver.error(m);
}
